run_metadata:
  default_input_namespace: {{instance._local_config['namespaces']['input'] if instance._local_config['namespaces']['input'] else 'null'}}
  default_output_namespace: {{instance._local_config['namespaces']['output']}}
  description: Project description
  script: echo "Hello World!"
