#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Parse User Config
=================

Perform parsing of the user updated `config.yaml` file.


Contents
========

Modules
-------

    globbing - swap glob expressions for registry entries
    variables  - substitute recognised FAIR CLI variables 

"""

__date__ = "2021-08-04"




